package com.archeus.dex.xml.run;

import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;

import org.xml.sax.InputSource;

import com.archeus.FormatedText;
import com.archeus.FormatedTextHelper;
import com.archeus.FormatedWord;
import com.archeus.FormatedWordList;
import com.archeus.WordList;
import com.archeus.WordListArray;
import com.archeus.dex.html.HtmlParser;
import com.archeus.dex.xml.exceptions.InvalidUsage;

/**
 * <p>
 * Clasa are rolul de a genera XML-ul aferent unei definitii din DEX.
 * </p>
 * <p>
 * Va servi ca model de <i>cum se apeleaza face procesarea</i>.
 * </p>
 * 
 * @author ionut
 * @since Feb 18, 2007 9:28:00 PM
 */
public class DEXtoXML {

	public static String[] pos = {
			"s", // subst
			"m", "f",
			"n", // masc, fem, neutru
			"vb", // verb
			"I", "II", "III", "IV", "V"/* ?? */, "intranz", "tranz", "adv",
			"adj", "conj", "prep", "interj", };

	public static String[] sense1 = { "A", "B", "C", "D", "E", "F", "G", "H",
	// "I", "J", "K", "L", "M", "N", "O", "P", "R", "S", "T", "U", "V",
	};

	public static String[] sense2 = { "I", "II", "III", "IV", "V", "VI", "VII",
			"VIII", "IX", "X", "XI", "XII", "XIII", "XIV", "XV", };

	public static String[] sense3 = { "1", "2", "3", "4", "5", "6", "7", "8",
			"9", "10", "11", "12", "13", "14", "15", "16", "17", "18", "19",
			"20", "21", "22", "23", "24", "25", };

	/**
	 * @author ionut
	 * @since Feb 18, 2007 9:27:56 PM
	 * @param args -
	 *            numele fisierului de intrare si eventul al celui de iesire.
	 *            Daca nu este specificat fisier de iesire, se va asifa la
	 *            consola. Fisierul de intrare trebuie sa fie in format html,
	 *            asa cum apare el in baza de date de la dexonline
	 */
	public static void main(String[] args) throws Exception {
		if (args.length == 0)
			throw new InvalidUsage("Specificati fisierul sursa!");
		File f = new File(args[0]);
		if (!f.exists())
			throw new InvalidUsage("Fisierul sursa nu exista!");
		try {
			InputSource src = new InputSource(new FileReader(f));
			src.setEncoding("UTF-8");
			WordListArray res = new WordListArray();
			parseDefinition(src, res);

		} catch (Exception e) {
			e.printStackTrace();
		}

	}

	public static void parseDefinition(InputSource src, WordListArray res)
			throws Exception {
		ArrayList<FormatedText> text = HtmlParser.parseFile(src);
		FormatedWordList list = FormatedTextHelper.wordList(text);
		int idxCrt = getHead(list, 0, res);
		int idxNext = getFeminine(list, idxCrt, res);
		// se cauta 'perechi' plural-pos, unde pluralul poate lipsi
		for (;;) {
			// se cauta plural
			WordList g = new WordList(WordList.Type.PLURAL_GROUP);
			idxNext = getPlural(list, idxNext == -1 ? idxCrt : idxNext, g
					.getChildren());
			if (idxNext != -1)
				idxCrt = idxNext;
			// se cauta pos chiar daca nu s-a gasit plural
			idxNext = getPOS(list, idxCrt, g.getChildren());
			if (g.getChildren().size() > 0)
				res.add(g);
			if (idxNext == -1)
				break;
			idxCrt = idxNext;
		}
		// sensuri
		WordList s = new WordList(WordList.Type.SENSES);
		res.add(s);
		// se determina nivelul
		idxCrt = skipBlank(list, idxCrt, res.get(res.size() - 1));
		if ((list.size() > idxCrt) ){//&& (isSenseNumber(list, idxCrt))) {
			int level = getSenseLevel(list.get(idxCrt).getText());
			for (;;) {
				WordList w = new WordList(WordList.Type.SENSE);
				int oldIdx = idxCrt;
				idxCrt = getSense(list, idxCrt, w, level);
				if (oldIdx == idxCrt)
					break;
				s.getChildren().add(w);
				oldIdx = idxCrt;
			}
		}
		// provenien ta
		if (isProvenance(list, idxCrt)) {
			WordList w = new WordList(WordList.Type.PROVENANCE);
			res.add(w);
			while (idxCrt < list.size())
				w.getList().add(list.get(idxCrt++));
		} // if provenienta
	}

	public static void checkIndex(FormatedWordList list, int idx)
			throws Exception {
		if (idx >= list.size())
			throw new Exception("index out of bounds: " + idx + "/"
					+ list.size());
	}

	/**
	 * <p>
	 * Intoarce indexul primului cuvant diferit de spatiu
	 * </p>
	 * 
	 * @author ionut
	 * @since Feb 25, 2007 11:12:54 AM
	 * @param list
	 * @param idx
	 * @return
	 * @throws Exception
	 */
	public static int skipBlank(FormatedWordList list, int idx, WordList dest)
			throws Exception {
		// checkIndex(list, idx);
		if (idx >= list.size())
			return idx;
		while (list.get(idx).getText().equals(" ")) {
			if (dest != null)
				dest.getList().add(list.get(idx++));
			if (list.size() == idx)
				break;
		} // while
		return idx;
	}

	/**
	 * <p>
	 * Determina <code>head</code>-ul unei definitii.
	 * </p>
	 * <p>
	 * <code>Head</code>-ul este prima parte a definitiei, marcata cu
	 * <code>bold</code>
	 * </p>
	 * <p>
	 * Este un singur cuvant, deci nu poate contine spatii; se termina cu
	 * virgula
	 * </p>
	 * 
	 * @author ionut
	 * @since Feb 25, 2007 10:35:19 AM
	 * @param list
	 * @param idx
	 * @param res
	 * @return
	 * @throws Exception
	 */
	public static int getHead(FormatedWordList list, int idx, WordListArray res)
			throws Exception {
		checkIndex(list, idx);
		if (!list.get(idx).isBold())
			throw new Exception("<bold> expected");
		// se considera head primul cuvat (inclusiv o eventuala virgula)
		WordList w = new WordList(WordList.Type.HEAD);
		res.add(w);
		w.getList().add(list.get(idx++));
		checkIndex(list, idx);
		// if (list.get(idx).isBold()&&(list.get(idx).getText().equals(",")))
		if (list.get(idx).getText().equals(","))
			w.getList().add(list.get(idx++));
		return idx;
	}

	/**
	 * <p>
	 * Partea de feminin. Bold, incepe cu '-' si te incheie cu ','
	 * </p>
	 * 
	 * @author ionut
	 * @since Feb 25, 2007 1:25:27 PM
	 * @param list
	 * @param idx
	 * @param res
	 * @return
	 * @throws Exception
	 */
	public static int getFeminine(FormatedWordList list, int idx,
			WordListArray res) throws Exception {
		checkIndex(list, idx);
		WordList w = new WordList(WordList.Type.FEMININE);
		idx = skipBlank(list, idx, w);
		if (!list.get(idx).getText().equals("-"))
			return idx;
		// se adauga '-'
		w.getList().add(list.get(idx++));
		// se adauga forma de plural
		checkIndex(list, idx);
		w.getList().add(list.get(idx++));
		// se adauga o eventuala virgula
		checkIndex(list, idx);
		if (list.get(idx).getText().equals(","))
			w.getList().add(list.get(idx++));
		res.add(w);
		return idx;
	}

	/**
	 * <p>
	 * Pluralul este imediat dupa head, poate lipsi.
	 * </p>
	 * <p>
	 * Este marcat cu <code>italic</code>, uneori precedat de numarul
	 * definitii intre paranteze rotunde (vezi definitia de la <code>timp</code>.)
	 * </p>
	 * 
	 * @author ionut
	 * @since Feb 25, 2007 11:07:46 AM
	 * @param list
	 * @param idx
	 * @param res
	 * @return
	 * @throws Exception
	 */
	public static int getPlural(FormatedWordList list, int idx,
			WordListArray res) throws Exception {
		boolean found = false;
		WordList w = new WordList(WordList.Type.PLURAL);
		idx = skipBlank(list, idx, w);
		checkIndex(list, idx);

		if (list.get(idx).getText().equals("(")) {
			WordList wc = new WordList(WordList.Type.PLURAL_COMMENT);
			w.getChildren().add(wc);
			// se "sare" la ')'
			while (list.size() > idx) {
				wc.getList().add(list.get(idx));
				if (list.get(idx++).getText().equals(")"))
					break;
			} // while
		} // is '('
		idx = skipBlank(list, idx, w);
		// se cauta un cuvant italic (sau mai multe)
		if (list.get(idx).isItalic()) {
			while (list.get(idx).isItalic()
					|| list.get(idx).getText().equals(",")) {
				w.getList().add(list.get(idx++));
			} // while italic
			found = true;
			res.add(w);
		} // if italic

		return found ? idx : -1;
	}

	/**
	 * <p>
	 * Deterimina daca <code>s</code> este parte de vorbire (abreviere)
	 * </p>
	 * 
	 * @author ionut
	 * @since Feb 25, 2007 2:14:44 PM
	 * @param s
	 * @return
	 */
	public static boolean isPOS(String s) {
		for (String p : pos)
			if (s.toUpperCase().equals(p.toUpperCase()))
				return true;
		return false;
	}

	/**
	 * <p>
	 * Deterimina daca <code>s</code> este numar sens
	 * </p>
	 * 
	 * @author ionut
	 * @since Feb 25, 2007 2:14:44 PM
	 * @param s
	 * @return
	 */
	public static boolean isSenseNumber(FormatedWordList list, int idx) {
		if (idx + 1 >= list.size())
			return false;
		if (getSenseLevel(list.get(idx).getText()) == -1)
			return false;
		if (list.get(idx + 1).getText().equals("."))
			return true;
		return false;
	}

	/**
	 * <p>
	 * Determina <code>nivelul</code>sensului
	 * </p>
	 * 
	 * @author ionut
	 * @since Feb 25, 2007 11:08:28 PM
	 * @param s
	 * @return -1 daca nu este numar sens
	 */
	public static int getSenseLevel(String s) {
		for (String p : sense1)
			if (s.equals(p))
				return 1;
		for (String p : sense2)
			if (s.equals(p))
				return 2;
		for (String p : sense3)
			if (s.equals(p))
				return 3;
		return -1;
	}

	/**
	 * <p>
	 * Partera de vorbire este formata din prescurtari (<code>pos</code>
	 * urmat de punct)
	 * </p>
	 * 
	 * @author ionut
	 * @since Feb 25, 2007 2:05:49 PM
	 * @param list
	 * @param idx
	 * @param res
	 * @return
	 * @throws Exception
	 */
	public static int getPOS(FormatedWordList list, int idx, WordListArray res)
			throws Exception {
		boolean found = false;
		checkIndex(list, idx);
		WordList w = new WordList(WordList.Type.POS);
		for (;;) {
			idx = skipBlank(list, idx, w);
			if (list.size() <= idx + 2)
				break;
			FormatedWord w1 = list.get(idx);
			FormatedWord w2 = list.get(idx + 1);
			if (w2.getText().equals(".") && isPOS(w1.getText())) {
				if (!(isSenseNumber(list, idx) && w1.isBold())) {
					w.getList().add(w1);
					w.getList().add(w2);
					found = true;
					idx = idx + 2;
					// se ia si virgula, dca exista
					idx = skipBlank(list, idx, w);
					if ((idx < list.size())
							&& list.get(idx).getText().equals(","))
						w.getList().add(list.get(idx++));
					continue;
				} // if bold
			} // if pos
			else {
				// tratare caz special: S.m. si (rar) n.
				// nu se modifica idx decat dupa ce suntem siguri ca avem cazul
				// cautat
				int i = idx;
				String s = w1.getText().toUpperCase();
				if ((list.size() > idx + 3)
						&& (s.equals("SI") || s.equals("ŞI") || s
								.equals("%C5%9EI")) && w2.getText().equals(" ")) {
					WordList tmp = new WordList(WordList.Type.POS);
					tmp.getList().add(w1);
					tmp.getList().add(w2);
					i = idx + 2;
					w2 = list.get(idx + 2);
					if (w2.getText().equals("(")) {
						// trebuie gasita ")"
						// cazuri cunoscute: rar, înv.
						while (list.size() > i) {
							tmp.getList().add(list.get(i));
							if (list.get(i++).getText().equals(")"))
								break;
						} // while
						i = skipBlank(list, i, tmp);
						if (list.size() > i + 1) {
							if (isPOS(list.get(i).getText())
									&& list.get(i + 1).getText().equals(".")) {
								// este cazul special cautat
								tmp.getList().add(list.get(i));
								tmp.getList().add(list.get(i + 1));
								for (FormatedWord item : tmp.getList())
									w.getList().add(item);
								idx = skipBlank(list, i + 2, w);
								continue;
							} // if
						}
					} // if "("
				} // if "SI"
			}// else
			break;

		} // for
		if (found)
			res.add(w);
		return found ? idx : -1;
	}

	/**
	 * <p>
	 * Determina daca la pozitia curenta se gasesc informatii despre provenienta
	 * </p>
	 * 
	 * @author ionut
	 * @since Feb 25, 2007 11:40:37 PM
	 * @param list
	 * @param idx
	 * @return
	 */
	public static boolean isProvenance(FormatedWordList list, int idx) {
		String[] prov = {"Din", "Lat"};
		if (list.size() <= idx + 3)
			return false;
		String crt = list.get(idx).getText();
		// cazul in care a fost omis spatiul
		for (String s : prov) {
			if (crt.equals("-"+s)) 
				return true;
		}// for
		if (!crt.equals("-"))
			return false;
		if (!list.get(idx + 1).getText().equals(" "))
			return false;
		crt = list.get(idx + 2).getText();
		for (String s : prov) {
			if (crt.equals(s)) 
				return true;
		}// for
		return false;
	}

	/**
	 * <p>
	 * Se cauta definitii de nivel <code>level</code>. Daca se gaseste un
	 * nivel superior se preda controlul la apeleant. Daca se gaseste un nivel
	 * inferior se proceseaza recursiv.
	 * </p>
	 * 
	 * @author ionut
	 * @since Feb 25, 2007 11:35:43 PM
	 * @param list
	 * @param idx
	 * @param res
	 * @param level
	 * @return
	 * @throws Exception
	 */
	public static int getSense(FormatedWordList list, int idx, WordList res,
			int level) throws Exception {
		if (isProvenance(list, idx))
			return idx;
		if (idx >= list.size())
			return idx;
		if (!isSenseNumber(list, idx)) {
			// nu este sens, se considera explicatie
			int idxPOs = getPOS(list, idx, res.getPos());
			if (idxPOs != -1)
				idx = idxPOs;
			boolean expresii = false;
			while ((idx < list.size()) && !isSenseNumber(list, idx)
					&& !isProvenance(list, idx)) {
				String crt = list.get(idx).getText();
				expresii = expresii || (crt.equals("*")||crt.equals("**")||crt.equals(String.valueOf((char)4)));
				if (expresii)
					res.getExp().add(list.get(idx));
				else
					res.getList().add(list.get(idx));
				idx++;
			}
			// se cauta in adancime
			for (;;) {
				WordList w = new WordList(WordList.Type.SENSE);
				int oldIdx = idx;
				idx = getSense(list, idx, w, level);
				if (oldIdx == idx)
					break;
				res.getChildren().add(w);
				oldIdx = idx;
			}
			return idx;
		} // if !isSensNumber
		int crtLevel = getSenseLevel(list.get(idx).getText());
		if (crtLevel < level) {
			// trebuie preadat controlul la apelant
			return idx;
		} else if (crtLevel > level) {
			for (;;) {
				WordList w = new WordList(WordList.Type.SENSE);
				int oldIdx = idx;
				idx = getSense(list, idx, w, crtLevel);
				if (oldIdx == idx)
					break;
				res.getChildren().add(w);
				oldIdx = idx;
			}
			return idx;
		} else {
			// sunt pe acelasi nivel
			// se formeaza un nod cu sensul si la acesta se adauga definitiile
			if (!res.getName().equals(WordList.Type.SENSE))
				throw new Exception("WordList.Type.SENSE = " + res.getName());
			res.setName(list.get(idx).getText());
			idx = idx + 2; // numar sens,punct
			idx = skipBlank(list, idx, res);

			// se cauta in adancime
			if (isSenseNumber(list, idx))
				for (;; isSenseNumber(list, idx)) {
					WordList w = new WordList(WordList.Type.SENSE);
					int oldIdx = idx;
					idx = getSense(list, idx, w, level + 1);
					if (oldIdx == idx)
						break;
					res.getChildren().add(w);
					oldIdx = idx;
				}
			else {
				idx = getSense(list, idx, res, crtLevel + 1);
				idx = skipBlank(list, idx, res);
			}
			return idx;
		}
	}

}
